from asyncio import Lock
from collections import namedtuple
from contextlib import contextmanager
from datetime import datetime
from typing import TYPE_CHECKING, AsyncIterator, Iterator, Literal, overload

import nio
import sqlalchemy as sa
from sqlalchemy import orm

if TYPE_CHECKING:
    from matridge.matrix import Client

ReactionTarget = namedtuple("ReactionTarget", ["room", "event"])


class Base(orm.DeclarativeBase):
    pass


class Room(Base):
    __tablename__ = "rooms"
    __table_args__ = (sa.Index("rooms_room_id", "room_id", unique=True),)

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    room_id: orm.Mapped[str] = orm.mapped_column()

    messages: orm.Mapped[list["Message"]] = orm.relationship(
        "Message", back_populates="room"
    )


class Message(Base):
    __tablename__ = "messages"
    __table_args__ = (
        sa.Index("messages_event_id", "room_id", "event_id", unique=True),
    )

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    event_id: orm.Mapped[str] = orm.mapped_column()
    added: orm.Mapped[datetime] = orm.mapped_column(default=sa.func.now())

    room_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("rooms.id"))

    room: orm.Mapped[Room] = orm.relationship("Room", back_populates="messages")
    reactions: orm.Mapped[list["Reaction"]] = orm.relationship(
        "Reaction", back_populates="message", cascade="all, delete-orphan"
    )


class Reaction(Base):
    __tablename__ = "reactions"
    __table_args__ = (
        sa.Index("reactions_event_id", "event_id", "message_id", unique=True),
    )

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    event_id: orm.Mapped[str] = orm.mapped_column()
    sender: orm.Mapped[str] = orm.mapped_column()
    emoji: orm.Mapped[str] = orm.mapped_column()

    message_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("messages.id"))
    message: orm.Mapped[Message] = orm.relationship(
        "Message", back_populates="reactions"
    )


class ReactionCache:
    """
    To avoid fetching history on each matrix reaction event, we store the
    "reaction state" per message.

    This is because matrix reaction events are atomic, unlike XMPP reactions
    which contain the full state in each event.
    """

    def __init__(self, client: "Client"):
        self.matrix = client
        self.log = client.session.log

    @contextmanager
    def session(self) -> Iterator[orm.Session]:
        with Session() as session:
            yield session

    async def _fetch_if_needed(
        self, session: orm.Session, target: ReactionTarget
    ) -> tuple[Message, bool]:
        # we use a lock to prevent parallel calls to _fetch for the same target
        if target in _locks:
            del_lock = False
            lock = _locks[target]
        else:
            del_lock = True
            lock = _locks[target] = Lock()
        async with lock:
            room = session.scalar(sa.select(Room).filter_by(room_id=target.room))

            if room is None:
                room = Room(room_id=target.room)
                session.add(room)

            message = session.scalar(
                sa.select(Message).filter_by(room=room, event_id=target.event)
            )
            fetch = message is None
            if fetch:
                message = Message(room=room, event_id=target.event)
                session.add(message)
                async for reaction in self._fetch(target.room, target.event):
                    reaction.message = message
                    session.add(reaction)
        if del_lock:
            del _locks[target]
        assert message is not None
        return message, fetch

    async def _fetch(self, room: str, event_id: str) -> AsyncIterator[Reaction]:
        self.log.debug("fetching reactions for %s", event_id)
        async for event in self.matrix.room_get_event_relations(
            room, event_id, nio.api.RelationshipType.annotation
        ):
            if not isinstance(event, nio.ReactionEvent):
                self.log.warning("got a non-reaction event: %s", event)
                continue
            if not event.reacts_to == event_id:
                self.log.warning(
                    "request reaction on %s but got reaction on %s",
                    event_id,
                    event.reacts_to,
                )
                continue
            yield Reaction(
                event_id=event.event_id, sender=event.sender, emoji=event.key
            )

    async def add(
        self,
        session: orm.Session,
        room: str,
        msg: str,
        sender: str,
        emoji: str,
        reaction_event: str,
    ) -> None:
        target = ReactionTarget(room=room, event=msg)
        message, fetched = await self._fetch_if_needed(session, target)
        if fetched:
            # no need to add the reaction if we fetched, because we fetched the reaction
            # we want to add already
            return
        reaction = Reaction(
            event_id=reaction_event, sender=sender, emoji=emoji, message=message
        )
        session.add(reaction)

    @overload
    async def get(
        self,
        session: orm.Session,
        room: str,
        msg: str,
        sender: str,
        with_event_ids: Literal[False],
    ) -> set[str]: ...

    @overload
    async def get(
        self, session: orm.Session, room: str, msg: str, sender: str
    ) -> set[str]: ...

    @overload
    async def get(
        self,
        session: orm.Session,
        room: str,
        msg: str,
        sender: str,
        with_event_ids: Literal[True],
    ) -> dict[str, str]: ...

    async def get(
        self,
        session: orm.Session,
        room: str,
        msg: str,
        sender: str,
        with_event_ids=False,
    ):
        message, _ = await self._fetch_if_needed(
            session, ReactionTarget(room=room, event=msg)
        )
        stmt = sa.select(Reaction).filter_by(message=message, sender=sender)
        reactions = session.scalars(stmt).all()

        if with_event_ids:
            return {r.emoji: r.event_id for r in reactions}
        else:
            return set(r.emoji for r in reactions)

    @staticmethod
    def remove(
        session: orm.Session, room_id: str, event_id: str
    ) -> ReactionTarget | None:
        room = session.scalar(sa.select(Room).filter_by(room_id=room_id))
        if room is None:
            return None

        stmt = (
            sa.select(Reaction)
            .options(orm.joinedload(Reaction.message))
            .join(Reaction.message)
            .filter(Message.room_id == room.id, Reaction.event_id == event_id)
        )
        reaction = session.scalar(stmt)
        if reaction is None:
            return None
        session.delete(reaction)
        return ReactionTarget(room=room.room_id, event=reaction.message.event_id)


def purge_old_messages(limit: int):
    with Session() as session:
        current_message_count = session.query(Message).count()

        if current_message_count <= limit:
            return

        stmt = sa.delete(Message).where(
            Message.id.in_(
                sa.select(Message.id)
                .order_by(Message.added)
                .limit(current_message_count - limit)
            )
        )
        session.execute(stmt)
        session.commit()


__all__ = ("ReactionCache", "purge_old_messages")

_locks = dict[ReactionTarget, Lock]()

engine = sa.create_engine("sqlite:///:memory:")
Base.metadata.create_all(engine)
Session = orm.sessionmaker(engine)



import io
import json
from asyncio import Lock
from typing import Any, Iterable, Optional, Union

import aiohttp
import nio
from slidge import BaseSession, global_config
from slidge.util.types import (
    LegacyMessageType,
    LegacyThreadType,
    LinkPreview,
    Mention,
    PseudoPresenceShow,
    ResourceDict,
)
from slixmpp.exceptions import XMPPError

from .contact import Contact, Roster
from .group import MUC, Bookmarks, Participant
from .matrix import Client
from .util import get_content

Sender = Union[Contact, Participant]
Recipient = Union[MUC, Contact]


class Session(BaseSession[str, Recipient]):
    bookmarks: Bookmarks
    contacts: Roster

    MESSAGE_IDS_ARE_THREAD_IDS = True

    def __init__(self, *a):
        super().__init__(*a)
        self.events_to_ignore = set[str]()
        self.migrate_shelf()
        self.send_lock = Lock()
        f = self.user.legacy_module_data
        self.matrix = Client(f["homeserver"], f["user_id"], self)  # type:ignore

    def migrate_shelf(self):
        store_path = global_config.HOME_DIR / self.user_jid.bare
        if not store_path.exists():
            return
        user = self.user
        with store_path.open("r") as f:
            user.legacy_module_data = json.load(f)
        self.xmpp.store.users.update(user)
        self.log.info(
            "Credentials info transferred from %s to the slidge DB", store_path
        )
        store_path.unlink()

    async def login(self):
        await self.matrix.login_token()
        await self.matrix.listen()
        self.contacts.user_legacy_id = self.matrix.user_id  # type:ignore
        return f"Logged in as {self.matrix.user}"

    async def logout(self):
        self.matrix.stop_listen()

    async def __relates_to(
        self,
        room_id: str,
        content: dict[str, Any],
        reply_to_msg_id: Optional[str],
        thread: Optional[str],
    ):
        relates_to = dict[str, Any]()
        if reply_to_msg_id:
            relates_to["m.in_reply_to"] = {
                "event_id": await self.matrix.get_original_id(room_id, reply_to_msg_id)
            }
        if thread:
            relates_to["rel_type"] = "m.thread"
            relates_to["event_id"] = thread
        if relates_to:
            content["m.relates_to"] = relates_to

    async def __handle_response(self, response: nio.Response):
        self.log.debug("Send response: %s", response)
        if isinstance(response, nio.RoomSendError):
            raise XMPPError("internal-server-error", str(response))
        assert isinstance(response, nio.RoomSendResponse)
        i = response.event_id
        self.events_to_ignore.add(i)
        return i

    async def __room_send(
        self, chat: MUC, content: dict, message_type="m.room.message"
    ):
        await self.matrix.room_typing(chat.legacy_id, False)
        async with self.send_lock:
            response = await self.matrix.room_send(
                chat.legacy_id,
                message_type=message_type,
                content=content,
                ignore_unverified_devices=self.user.preferences.get(
                    "trust_everything", False
                ),
            )
            return await self.__handle_response(response)

    async def on_text(
        self,
        chat: Recipient,
        text: str,
        *,
        reply_to_msg_id: Optional[str] = None,
        reply_to_fallback_text: Optional[str] = None,
        reply_to: Optional[Sender] = None,  # type: ignore
        thread: Optional[str] = None,  # type: ignore
        link_previews: Iterable[LinkPreview] = (),
        mentions: Optional[list[Mention]] = None,
    ) -> Optional[LegacyMessageType]:
        if isinstance(chat, Contact):
            raise XMPPError("bad-request", "Matridge does not implement 1:1 chats")
        content = get_content(text, mentions)
        await self.__relates_to(chat.legacy_id, content, reply_to_msg_id, thread)
        return await self.__room_send(chat, content)

    async def on_file(
        self,
        chat: Recipient,
        url: str,
        *,
        http_response: aiohttp.ClientResponse,
        reply_to_msg_id: Optional[str] = None,
        reply_to_fallback_text: Optional[str] = None,
        reply_to: Optional[Sender] = None,  # type: ignore
        thread: Optional[str] = None,  # type: ignore
    ) -> Optional[LegacyMessageType]:
        if isinstance(chat, Contact):
            raise XMPPError("bad-request", "Matridge does not implement 1:1 chats")
        filename = url.split("/")[-1]
        content_type = http_response.content_type
        resp, _ = await self.matrix.upload(
            io.BytesIO(await http_response.read()),
            content_type,
            filename,
            filesize=http_response.content_length,
        )
        self.log.debug("Upload response: %s %r", type(resp), resp)
        if not isinstance(resp, nio.UploadResponse):
            raise XMPPError("internal-server-error", str(resp))
        content = {
            "msgtype": "m.image" if content_type.startswith("image") else "m.file",
            "body": filename,
            "url": resp.content_uri,
        }
        await self.__relates_to(chat.legacy_id, content, reply_to_msg_id, thread)
        return await self.__room_send(chat, content)

    async def on_composing(
        self, chat: Recipient, thread: Optional[LegacyThreadType] = None
    ):
        if isinstance(chat, Contact):
            raise XMPPError("bad-request", "Matridge does not implement 1:1 chats")
        await self.matrix.room_typing(chat.legacy_id)

    async def on_paused(
        self, chat: Recipient, thread: Optional[LegacyThreadType] = None
    ):
        if isinstance(chat, Contact):
            raise XMPPError("bad-request", "Matridge does not implement 1:1 chats")
        await self.matrix.room_typing(chat.legacy_id, False)

    async def on_displayed(
        self,
        chat: Recipient,
        legacy_msg_id: LegacyMessageType,
        thread: Optional[LegacyThreadType] = None,
    ):
        if isinstance(chat, Contact):
            raise XMPPError("bad-request", "Matridge does not implement 1:1 chats")
        resp = await self.matrix.update_receipt_marker(chat.legacy_id, legacy_msg_id)
        self.log.debug("Displayed response: %s", resp)

    async def on_correct(
        self,
        chat: Recipient,
        text: str,
        legacy_msg_id: str,
        *,
        thread: Optional[str] = None,  # type: ignore
        link_previews: Iterable[LinkPreview] = (),
        mentions: Optional[list[Mention]] = None,
    ) -> Optional[str]:
        if isinstance(chat, Contact):
            raise XMPPError("bad-request", "Matridge does not implement 1:1 chats")
        content = {
            "msgtype": "m.text",
            "body": "* " + text,
            "m.new_content": get_content(text, mentions),
            "m.relates_to": {"rel_type": "m.replace", "event_id": legacy_msg_id},
        }
        await self.__relates_to(chat.legacy_id, content, None, thread)
        return await self.__room_send(chat, content)

    async def on_react(
        self,
        chat: Recipient,
        legacy_msg_id: str,
        emojis: list[str],
        thread: Optional[LegacyThreadType] = None,
    ):
        if isinstance(chat, Contact):
            raise XMPPError("bad-request", "Matridge does not implement 1:1 chats")
        new_emojis = set(emojis)
        with self.matrix.reactions.session() as session:
            old_emojis = await self.matrix.reactions.get(
                session,
                chat.legacy_id,
                legacy_msg_id,
                self.matrix.user_id,
                with_event_ids=True,
            )
            for old_emoji, event in old_emojis.items():
                if old_emoji in new_emojis:
                    new_emojis.remove(old_emoji)
                else:
                    await self.on_retract(chat, event)
                    self.matrix.reactions.remove(session, chat.legacy_id, event)

            for emoji in new_emojis:
                content = {
                    "m.relates_to": {
                        "rel_type": "m.annotation",
                        "event_id": legacy_msg_id,
                        "key": emoji,
                    },
                }

                i = await self.__room_send(chat, content, "m.reaction")
                await self.matrix.reactions.add(
                    session,
                    chat.legacy_id,
                    legacy_msg_id,
                    self.matrix.user_id,
                    emoji,
                    i,
                )
            session.commit()

    async def on_retract(
        self,
        chat: Recipient,
        legacy_msg_id: str,
        thread: Optional[str] = None,  # type: ignore
    ):
        if isinstance(chat, Contact):
            raise XMPPError("bad-request", "Matridge does not implement 1:1 chats")
        resp = await self.matrix.room_redact(chat.legacy_id, legacy_msg_id)
        self.log.debug("Redact response: %s", resp)
        if isinstance(resp, nio.RoomRedactError):
            raise XMPPError("internal-server-error", str(resp))
        assert isinstance(resp, nio.RoomRedactResponse)
        self.events_to_ignore.add(resp.event_id)

    async def on_presence(
        self,
        resource: str,
        show: PseudoPresenceShow,
        status: str,
        resources: dict[str, ResourceDict],
        merged_resource: Optional[ResourceDict],
    ):
        if not merged_resource:
            resp = await self.matrix.set_presence("offline")
        else:
            resp = await self.matrix.set_presence(
                PRESENCE_DICT[merged_resource["show"]],
                merged_resource["status"] or None,
            )
        self.log.debug("Set presence response: %s", resp)

    async def on_avatar(
        self,
        bytes_: Optional[bytes],
        hash_: Optional[str],
        type_: Optional[str],
        width: Optional[int],
        height: Optional[int],
    ) -> None:
        if bytes_:
            resp, _ = await self.matrix.upload(
                io.BytesIO(bytes_), type_, filesize=len(bytes_)
            )
            self.log.debug("Upload response: %s %r", type(resp), resp)
            if not isinstance(resp, nio.UploadResponse):
                raise XMPPError("internal-server-error", str(resp))
            uri = resp.content_uri
        else:
            uri = ""
        await self.matrix.set_avatar(uri)

    async def on_leave_group(
        self,
        muc_legacy_id: str,  # type:ignore
    ):
        resp = await self.matrix.room_leave(muc_legacy_id)
        self.log.debug("Leave response: %s", resp)
        # This prevents the room being re-listed when the user calls the
        # "groups" command which in turn calls Bookmarks.fill().
        del self.matrix.rooms[muc_legacy_id]


PRESENCE_DICT: dict[PseudoPresenceShow, str] = {
    "": "online",
    "chat": "online",
    "away": "unavailable",
    "xa": "unavailable",
    "dnd": "unavailable",
}



import logging
import time
import warnings
from asyncio import Task, create_task, sleep
from functools import wraps
from typing import TYPE_CHECKING, AsyncIterator, Awaitable, Callable, Optional, Union

import nio
from async_lru import alru_cache
from nio import Event
from slidge.core import config as global_config
from slidge.util.types import LegacyAttachment
from slixmpp import JID
from slixmpp.exceptions import XMPPError

from .reactions import ReactionCache
from .util import Credentials, get_replace, server_timestamp_to_datetime

if TYPE_CHECKING:
    from .group import MUC, Participant
    from .session import Session


def catch_all(coro: Callable[["Client", nio.MatrixRoom, nio.Event], Awaitable[None]]):
    @wraps(coro)
    async def wrapped(self: "Client", room: nio.MatrixRoom, event: nio.Event, *a, **kw):
        event_id = getattr(event, "event_id", None)
        async with self.session.send_lock:
            if event_id in self.session.events_to_ignore:
                self.log.debug("Ignoring an event matridge has sent: %s", event_id)
                return
        try:
            return await coro(self, room, event, *a, **kw)
        except XMPPError as e:
            self.log.debug(
                "Exception raised in matrix client callback %s", coro, exc_info=e
            )
        except Exception as e:
            self.log.exception(
                "Exception raised in matrix client callback %s", coro, exc_info=e
            )

    return wrapped


class AuthenticationClient(nio.AsyncClient):
    def __init__(
        self, server: str, handle: str, jid: JID, log: Optional[logging.Logger] = None
    ):
        if not server.startswith("http"):
            server = "https://" + server
        self.store_path = store_path = global_config.HOME_DIR / (jid.bare + "_state")
        store_path.mkdir(exist_ok=True)
        cfg = nio.AsyncClientConfig(
            store_sync_tokens=True,
            max_limit_exceeded=0,
            max_timeouts=0,
            encryption_enabled=True,
        )
        super().__init__(server, handle, store_path=str(store_path), config=cfg)
        if log:
            self.log = log
        else:
            self.log = logging.getLogger(__name__)

    def get_credentials(self, resp: nio.LoginResponse) -> Credentials:
        return {
            "homeserver": self.homeserver,
            "user_id": resp.user_id,
            "device_id": resp.device_id,
            "access_token": resp.access_token,
        }

    async def fix_homeserver(self):
        """
        Uses https://$HOMESERVER/.well-known/matrix/client to fix the homeserver
        URL.
        """
        response = await self.discovery_info()
        if isinstance(response, nio.DiscoveryInfoResponse):
            self.homeserver = response.homeserver_url


class Client(AuthenticationClient):
    MIN_RETRY_TIME = 3
    MAX_RETRY_TIME = 300
    REQUEST_TIMEOUT = 30000
    CONSIDER_SUCCESSFUL = 10

    def __init__(self, server: str, handle: str, session: "Session"):
        super().__init__(server, handle, session.user.jid, session.log)
        self.__sync_task: Optional[Task] = None
        self.session = session
        self.reactions = ReactionCache(self)

    def load(self):
        stored = self.session.user.legacy_module_data
        self.access_token: str = stored["access_token"]  # type:ignore
        self.user_id: str = stored["user_id"]  # type:ignore
        self.device_id: str = stored["device_id"]  # type:ignore

    async def login_token(self):
        self.load()
        await self.fix_homeserver()
        self.load_store()

    def __add_event_handlers(self):
        self.add_event_callback(self.on_event, nio.Event)  # type:ignore
        self.add_event_callback(self.on_message, nio.RoomMessage)  # type:ignore
        self.add_event_callback(self.on_avatar, nio.RoomAvatarEvent)  # type:ignore
        self.add_event_callback(self.on_topic, nio.RoomTopicEvent)  # type:ignore
        self.add_event_callback(self.on_name, nio.RoomNameEvent)  # type:ignore
        self.add_event_callback(self.on_sticker, nio.StickerEvent)  # type:ignore
        self.add_event_callback(self.on_member, nio.RoomMemberEvent)  # type:ignore
        self.add_event_callback(self.on_redact, nio.RedactionEvent)  # type:ignore
        self.add_event_callback(self.on_reaction, nio.ReactionEvent)  # type:ignore
        self.add_event_callback(self.on_power_levels, nio.PowerLevelsEvent)  # type:ignore
        self.add_presence_callback(self.on_presence, nio.PresenceEvent)  # type:ignore
        self.add_ephemeral_callback(self.on_receipt, nio.ReceiptEvent)  # type:ignore
        self.add_ephemeral_callback(
            self.on_typing,  # type:ignore
            nio.TypingNoticeEvent,
        )

        self.add_to_device_callback(
            self.on_key_verification,  # type:ignore
            (nio.KeyVerificationEvent, nio.UnknownToDeviceEvent),
        )

    async def __get_muc(self, room: Union[nio.MatrixRoom, str]) -> "MUC":
        room_id = room.room_id if isinstance(room, nio.MatrixRoom) else room
        return await self.session.bookmarks.by_legacy_id(room_id)

    async def __sync_forever(self):
        attempts = 0
        while True:
            start = time.time()
            try:
                await self.sync_forever(timeout=self.REQUEST_TIMEOUT)
            except Exception as e:
                duration = time.time() - start
                if duration < self.CONSIDER_SUCCESSFUL:
                    attempts += 1
                    wait = min(attempts * self.MIN_RETRY_TIME, self.MAX_RETRY_TIME)
                else:
                    attempts = 0
                    wait = self.MIN_RETRY_TIME
                if attempts < 2:
                    self.log.debug(
                        "Sync task has raised %r, retrying in %s", e, wait, exc_info=e
                    )
                else:
                    self.log.error(
                        "Sync task has raised %r, retrying in %s", e, wait, exc_info=e
                    )
                await sleep(wait)
            else:
                break

    async def get_participant(
        self, room: nio.MatrixRoom, event: nio.Event
    ) -> "Participant":
        muc = await self.__get_muc(room)
        self.log.debug(
            "sender (%s) == me (%s)? %s",
            event.sender,
            self.session.contacts.user_legacy_id,
            event.sender == self.session.contacts.user_legacy_id,
        )
        return await muc.get_participant_by_legacy_id(event.sender)

    async def listen(self):
        # we need to sync full state or else we don't get the list of all rooms
        resp = await self.sync(full_state=True)
        self.log.debug("Sync")
        if isinstance(resp, nio.SyncError):
            raise PermissionError(resp)
        self.__add_event_handlers()
        self.__sync_task = create_task(self.__sync_forever())

    def stop_listen(self):
        if self.__sync_task is None:
            return
        self.__sync_task.cancel()

    async def fetch_history(self, room_id: str, limit: int) -> AsyncIterator[Event]:
        sync_resp = await self.sync()
        if isinstance(sync_resp, nio.SyncError):
            return
        resp = await self.room_messages(
            room_id,
            limit=limit,
            start=sync_resp.next_batch,
            message_filter={"types": ["m.room.message"]},
        )
        if not isinstance(resp, nio.RoomMessagesResponse):
            self.log.warning("Could not fill history.", sync_resp)
            return
        for event in resp.chunk:
            yield event

    @catch_all
    async def on_event(self, room: nio.MatrixRoom, event: nio.Event):
        self.log.debug("Event %s '%s': %r", type(event), room, event)

    @catch_all
    async def on_message(self, room: nio.MatrixRoom, event: nio.RoomMessage):
        self.log.debug("Message: %s", event)

        participant = await self.get_participant(room, event)
        await participant.send_matrix_message(event)

    async def on_presence(self, presence: nio.PresenceEvent):
        if presence.user_id == self.session.contacts.user_legacy_id:
            return
        try:
            contact = await self.session.contacts.by_legacy_id(presence.user_id)
        except XMPPError as e:
            self.log.debug("Ignoring presence: %s", presence, exc_info=e)
            return
        contact.update_presence(presence)

    @catch_all
    async def on_avatar(self, room: nio.MatrixRoom, event: nio.RoomAvatarEvent):
        muc = await self.__get_muc(room)
        muc.avatar = event.avatar_url

    @catch_all
    async def on_topic(self, room: nio.MatrixRoom, event: nio.RoomTopicEvent):
        muc = await self.__get_muc(room)
        participant = await self.get_participant(room, event)
        muc.subject = event.topic
        muc.subject_setter = participant
        muc.subject_date = server_timestamp_to_datetime(event)

    @catch_all
    async def on_name(self, room: nio.MatrixRoom, event: nio.RoomNameEvent):
        muc = await self.__get_muc(room)
        muc.name = event.name

    @catch_all
    async def on_sticker(self, room: nio.MatrixRoom, event: nio.StickerEvent):
        participant = await self.get_participant(room, event)

        resp = await self.download(event.url)
        if isinstance(resp, nio.DownloadResponse):
            await participant.send_files(
                [LegacyAttachment(data=resp.body, caption=event.body)]
            )
        else:
            self.log.error("Failed to download sticker: %r", resp)

    @catch_all
    async def on_member(self, room: nio.MatrixRoom, event: nio.RoomMemberEvent):
        muc = await self.__get_muc(room)
        participant = await self.get_participant(room, event)
        if event.membership == "join" and participant.contact:
            participant.set_affiliation_from_power_level(
                room.power_levels.get_user_level(participant.contact.legacy_id)
            )
        elif event.membership == "leave":
            if participant.is_user:
                await self.session.bookmarks.remove(muc)
            else:
                muc.remove_participant(participant)
        elif event.membership == "ban":
            # TODO: handle bans in slidge core
            pass
        elif event.membership == "invite":
            # TODO: what's that event exactly?
            pass

    @catch_all
    async def on_typing(self, room: nio.MatrixRoom, event: nio.TypingNoticeEvent):
        muc = await self.__get_muc(room)
        for user_id in event.users:
            participant = await muc.get_participant_by_legacy_id(user_id)
            participant.composing()

    @catch_all
    async def on_receipt(self, room: nio.MatrixRoom, event: nio.ReceiptEvent):
        muc = await self.__get_muc(room)
        for receipt in event.receipts:
            if receipt.receipt_type == "m.read":
                participant = await muc.get_participant_by_legacy_id(receipt.user_id)
                participant.displayed(receipt.event_id)

    @catch_all
    async def on_reaction(self, room: nio.MatrixRoom, event: nio.ReactionEvent, **kw):
        self.log.debug("Reaction2")

        source = event.source
        msg_id = await self.get_original_id(room.room_id, event.reacts_to)

        sender = source["sender"]
        emoji = event.key

        with self.reactions.session() as s:
            reactions = await self.reactions.get(
                s, room.room_id, msg_id, sender, with_event_ids=True
            )

        event_ids = list(reactions.values())
        emojis = list(reactions.keys())
        if event.event_id in event_ids:
            idx = event_ids.index(event.event_id)
            stored_emoji = emojis[idx]
            if stored_emoji != emoji:
                warnings.warn(
                    f"Received an emoji reaction update for the event ID '{event.event_id}': "
                    f"'{stored_emoji}' → '{emoji}'. This is illegal in Matrix, ignoring."
                )
        else:
            if emoji in emojis:
                warnings.warn(
                    f"We already had an event ID for the emoji reaction '{emoji}':"
                    f" '{event.event_id}' → {reactions[emoji]}."
                )
            with self.reactions.session() as s:
                await self.reactions.add(
                    s, room.room_id, msg_id, sender, emoji, event.event_id
                )
                s.commit()
            if emoji in reactions:
                warnings.warn(
                    f"The emoji '{emoji}' was already present with a different event ID"
                )
            else:
                emojis.append(emoji)

        participant = await self.get_participant(room, event)
        participant.react(msg_id, emojis, **kw)

    @catch_all
    async def on_redact(self, room: nio.MatrixRoom, event: nio.RedactionEvent):
        self.log.debug("Redaction: %s", event)
        redacter = await self.get_participant(room, event)
        with self.reactions.session() as s:
            if reaction_target := self.reactions.remove(s, room.room_id, event.redacts):
                msg_id = await self.get_original_id(room.room_id, reaction_target.event)
                reactions = await self.reactions.get(
                    s, reaction_target.room, msg_id, event.sender
                )
                redacter.react(msg_id, reactions)
                s.commit()
                return

        redacted_event = await self.get_event(room.room_id, event.redacts)

        if redacted_event and event.sender == redacted_event.sender:
            redacter.retract(event.redacts)
        else:
            redacter.moderate(event.redacts, event.reason)

    async def on_power_levels(self, room: nio.MatrixRoom, event: nio.PowerLevelsEvent):
        muc = await self.__get_muc(room)
        for user_id, power_level in event.power_levels.users.items():
            participant = await muc.get_participant_by_legacy_id(user_id)
            participant.set_affiliation_from_power_level(power_level)

    @alru_cache(maxsize=1000)
    async def get_original_id(self, room_id: str, event_id: str) -> str:
        event = await self.get_event(room_id, event_id)
        if event is None:
            return event_id
        # no need to check recursively because replacements must refer to
        # the original event
        return get_replace(event.source) or event_id

    @alru_cache(maxsize=100)
    async def get_event(self, room_id: str, event_id: str) -> Optional[nio.Event]:
        resp = await self.session.matrix.room_get_event(room_id, event_id)
        if isinstance(resp, nio.RoomGetEventError):
            return None
        return resp.event

    async def on_key_verification(
        self, event: nio.KeyVerificationEvent | nio.UnknownToDeviceEvent
    ):
        # TODO: once https://github.com/matrix-nio/matrix-nio/issues/430 is fixed, we should be able to simplify
        #       this by ditching UnknownToDeviceEvent
        # Thanks to DomNomNom from the nio Channel,
        # cf https://github.com/wreald/matrix-nio/commit/5cb8e99965bcb622101b1d
        content = event.source.get("content", {})
        if event.source.get("type") == "m.key.verification.request":
            self.session.send_gateway_message(
                "Got verification request. Waiting for other device to accept SAS method…"
            )
            methods = content.get("methods")
            if "m.sas.v1" not in methods:
                self.session.send_gateway_message(
                    "Received a device verification request, but other device does not support SAS authentication. "
                    f"Methods: {methods}."
                )
                return
            txid = content.get("transaction_id")
            ready_event = nio.ToDeviceMessage(
                type="m.key.verification.ready",
                recipient=event.sender,
                recipient_device=content.get("from_device"),
                content={
                    "from_device": self.device_id,
                    "methods": ["m.sas.v1"],
                    "transaction_id": txid,
                },
            )
            resp = await self.to_device(ready_event, txid)
            if isinstance(resp, nio.ToDeviceError):
                self.session.send_gateway_message(f"to_device failed with {resp}")
        elif isinstance(event, nio.KeyVerificationStart):  # first step
            if "emoji" not in event.short_authentication_string:
                self.session.send_gateway_message(
                    "Received a device verification request, but other device does not support emoji verification "
                    f"{event.short_authentication_string}."
                )
                return
            resp = await self.accept_key_verification(event.transaction_id)
            if isinstance(resp, nio.ToDeviceError):
                self.session.send_gateway_message(
                    f"accept_key_verification failed with {resp}"
                )

            sas = self.key_verifications.get(event.transaction_id)

            if sas is None:
                self.session.send_gateway_message("transaction_id could not be found")
                return

            todevice_msg = sas.share_key()
            resp = await self.to_device(todevice_msg)
            if isinstance(resp, nio.ToDeviceError):
                self.session.send_gateway_message(f"❌ to_device failed with {resp}")

        elif isinstance(event, nio.KeyVerificationCancel):  # anytime
            self.session.send_gateway_message(
                f"Verification has been cancelled by {event.sender} "
                f'for reason "{event.reason}".'
            )

        elif isinstance(event, nio.KeyVerificationKey):  # second step
            sas = self.key_verifications.get(event.transaction_id)

            if sas is None:
                self.session.send_gateway_message("transaction_id could not be found")
                return

            self.session.send_gateway_message(f"{sas.get_emoji()}")

            yn = await self.session.input("Do the emojis match? (Y/N) (C for Cancel) ")
            if yn.lower() == "y":
                resp = await self.confirm_short_auth_string(event.transaction_id)
                if isinstance(resp, nio.ToDeviceError):
                    self.session.send_gateway_message(
                        f"confirm_short_auth_string failed with {resp}"
                    )

                # Extra step in new flow: once we have completed the SAS
                # verification successfully, send a 'done' to-device event
                # to the other device to assert that the verification was
                # successful.
                done_message = nio.ToDeviceMessage(
                    type="m.key.verification.done",
                    recipient=event.sender,
                    recipient_device=sas.other_olm_device.device_id,
                    content={
                        "transaction_id": sas.transaction_id,
                    },
                )
                resp = await self.to_device(done_message, sas.transaction_id)
                if isinstance(resp, nio.ToDeviceError):
                    self.session.send_gateway_message(f"'done' failed with {resp}")

            elif yn.lower() == "n":  # no, don't match, reject
                resp = await self.cancel_key_verification(
                    event.transaction_id, reject=True
                )
                if isinstance(resp, nio.ToDeviceError):
                    self.session.send_gateway_message(
                        f"cancel_key_verification failed with {resp}"
                    )
            else:  # C or anything for cancel
                self.session.send_gateway_message(
                    "Cancelled by user! Verification will be cancelled."
                )
                resp = await self.cancel_key_verification(
                    event.transaction_id, reject=False
                )
                if isinstance(resp, nio.ToDeviceError):
                    self.session.send_gateway_message(
                        f"cancel_key_verification failed with {resp}"
                    )

        elif isinstance(event, nio.KeyVerificationMac):  # third step
            sas = self.key_verifications.get(event.transaction_id)

            if sas is None:
                self.session.send_gateway_message("transaction_id could not be found")
                return

            try:
                todevice_msg = sas.get_mac()
            except nio.LocalProtocolError as e:
                # e.g. it might have been cancelled by ourselves
                self.session.send_gateway_message(
                    f"Cancelled or protocol error: Reason: {e}.\n"
                    f"Verification with {event.sender} not concluded. "
                    "Try again?"
                )
            else:
                resp = await self.to_device(todevice_msg)
                if isinstance(resp, nio.ToDeviceError):
                    self.session.send_gateway_message(f"to_device failed with {resp}")
        elif event.source.get("type") == "m.key.verification.done":
            # Final step, other device acknowledges verification success.
            txid = content.get("transaction_id")
            sas = self.key_verifications[txid]

            self.session.send_gateway_message(
                f"sas.we_started_it = {sas.we_started_it}\n"
                f"sas.sas_accepted = {sas.sas_accepted}\n"
                f"sas.canceled = {sas.canceled}\n"
                f"sas.timed_out = {sas.timed_out}\n"
                f"sas.verified = {sas.verified}\n"
                f"sas.verified_devices = {sas.verified_devices}\n"
            )
            self.session.send_gateway_message("Emoji verification was successful!")
        else:
            self.session.send_gateway_message(
                f"Received unexpected event type {type(event)}. "
                f"Event is {event}. Event will be ignored."
            )



import asyncio
import logging
import warnings
from shutil import rmtree
from typing import TYPE_CHECKING, Optional

from bs4 import MarkupResemblesLocatorWarning
from nio.responses import LoginError
from slidge import BaseGateway, FormField, global_config
from slixmpp import JID
from slixmpp.exceptions import XMPPError

from . import config, reactions
from .matrix import AuthenticationClient
from .util import Credentials

if TYPE_CHECKING:
    from .session import Session


class Gateway(BaseGateway):
    REGISTRATION_FIELDS = [
        FormField(var="homeserver", label="Home Server", required=True),
        FormField(var="username", label="User name", required=True),
        FormField(var="password", label="Password", required=True, private=True),
        FormField(
            var="device",
            label="Device name",
            value=f"matridge on {getattr(global_config, 'JID', 'dev')}",
            required=True,
        ),
    ]
    REGISTRATION_INSTRUCTIONS: str = "Enter your credentials"

    COMPONENT_NAME = "Matrix (slidge)"
    COMPONENT_TYPE = "matrix"

    COMPONENT_AVATAR = (
        "https://codeberg.org/slidge/matridge/raw/branch/main/assets/matrix_outline.png"
    )

    ROSTER_GROUP: str = "matrix"

    MARK_ALL_MESSAGES = False

    PROPER_RECEIPTS = True

    GROUPS = True

    DB_PURGE_SLEEP = 3600 * 24

    PREFERENCES = BaseGateway.PREFERENCES + [
        FormField(
            var="trust_everything",
            label="Always trust new encryption keys. "
            "This is unsafe if you need perfect bridge to end encryption.",
            value="false",
            required=True,
            type="boolean",
        )
    ]

    def __init__(self):
        super().__init__()
        if not log.isEnabledFor(logging.DEBUG):
            logging.getLogger("nio.rooms").setLevel(logging.WARNING)
            logging.getLogger("nio.responses").setLevel(logging.ERROR)
            logging.getLogger("nio.events.misc").setLevel(logging.ERROR)
            logging.getLogger("nio.events.base_client").setLevel(logging.WARNING)
        self._db_purge_task = self.loop.create_task(self._db_purge())

    async def _db_purge(self):
        await asyncio.sleep(self.DB_PURGE_SLEEP)
        reactions.purge_old_messages(config.REACTION_CACHE_SIZE)

    async def validate(  # type:ignore[override]
        self, user_jid: JID, registration_form: dict[str, Optional[str]]
    ) -> Credentials:
        client = AuthenticationClient(
            registration_form["homeserver"],  # type:ignore
            registration_form["username"],  # type:ignore
            user_jid,
        )
        await client.fix_homeserver()
        resp = await client.login(
            registration_form["password"],  # type:ignore
            registration_form["device"],  # type:ignore
        )
        if isinstance(resp, LoginError):
            log.debug("Failed login: %r", resp)
            raise XMPPError("not-authorized", f"Could not login: {resp}")
        return client.get_credentials(resp)

    async def unregister(self, session: "Session"):  # type:ignore[override]
        await session.logout()
        rmtree(session.matrix.store_path)


log = logging.getLogger(__name__)
warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)



# GENERATED FILE, DO NOT EDIT DIRECTLY

# This file is automatically generated using https://codeberg.org/slidge/legacy-module-template/
# Its source is at:
# https://codeberg.org/slidge/legacy-module-template/src/branch/main/docs/source/conf/py.jinja
# Open a pull request for this repository instead, to benefit the documentation of all slidge-based gateways.

import os
from datetime import datetime
from pathlib import Path

from slidge import __version__ as slidge_version

project = "matridge"
copyright = f"{datetime.today().year}, the {project} contributors"
author = f"the {project} contributors"
branch = os.getenv("CI_COMMIT_BRANCH", "main")

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc.typehints",
    "slidge_sphinx_extensions.doap",
    "slidge_sphinx_extensions.config_obj",
    "sphinx_mdinclude",
]

autodoc_typehints = "description"

# Include __init__ docstrings
autoclass_content = "both"

extlinks = {"xep": ("https://xmpp.org/extensions/xep-%s.html", "XEP-%s")}

html_theme = "furo"
html_theme_options = {
    "source_edit_link": f"https://codeberg.org/slidge/{project}/_edit/{branch}/docs/source/{{filename}}",
    "source_view_link": f"https://codeberg.org/slidge/{project}/src/branch/{branch}/docs/source/{{filename}}",
    "footer_icons": [
        {
            "name": "Codeberg",
            "url": f"https://codeberg.org/slidge/{project}",
            "html": Path("codeberg.svg").read_text(),
        },
    ],
}


rst_epilog = f"""
.. |slidge_version| replace:: {slidge_version}
"""



import logging
import warnings
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Optional, TypedDict
from urllib.parse import quote

import bs4
import nio
from slidge.core.mixins import MessageMixin
from slidge.util.types import LegacyAttachment, Mention, MessageReference

try:
    from slidge_style_parser import format_for_matrix  # type:ignore
except ImportError as e:
    warnings.warn(f"Slidge style parser could not be imported: {e}")
    MESSAGE_STYLING = False
else:
    MESSAGE_STYLING = True
from slixmpp.exceptions import XMPPError

if TYPE_CHECKING:
    from .group import MUC
    from .session import Session


class MatrixMixin(MessageMixin):
    session: "Session"
    muc: "MUC"
    log: logging.Logger

    async def __get_reply_to(self, msg: nio.RoomMessage) -> Optional[MessageReference]:
        reply_to_msg_id = get_reply_to(msg.source)
        if not reply_to_msg_id:
            return None

        reply_to = MessageReference(legacy_id=reply_to_msg_id)
        if event := await self.muc.get_message(reply_to_msg_id):
            self.log.debug("Get Message Event: %r", event)
            try:
                author = await self.muc.get_participant_by_legacy_id(event.sender)
            except XMPPError as e:
                # maybe deleted profiles?
                self.log.debug(
                    "Something is wrong with participant %s, falling back to using matrix ID as nickname: %s",
                    event.sender,
                    e,
                )
                author = await self.muc.get_participant(event.sender)
            self.log.debug("Author: %r", author)
            reply_to.author = author
            if isinstance(event, nio.RoomMessage):
                reply_to.body = get_body(event)
        return reply_to

    async def __add_reply_to(self, msg: nio.RoomMessage, replace: str, kwargs: dict):
        if replace is None:
            kwargs["reply_to"] = await self.__get_reply_to(msg)
            return

        kwargs["correction"] = True
        original = await self.muc.get_message(replace)
        if not isinstance(original, nio.RoomMessage):
            self.log.warning(
                "Reply to something else than a message, or couldn't fetch it: %s",
                original,
            )
            return
        kwargs["reply_to"] = await self.__get_reply_to(original)

    async def __get_attachments(self, msg: nio.RoomMessage, **kwargs):
        if isinstance(msg, nio.RoomMessageMedia):
            return [
                LegacyAttachment(
                    url=await self.session.matrix.mxc_to_http(msg.url),
                    legacy_file_id=quote(msg.url),
                    name=get_body(msg) or None,
                )
            ]

        if isinstance(msg, nio.RoomEncryptedMedia):
            return await self.__get_encrypted_attachments(msg)

        # workaround for bots sending attachments that nio does not understand
        url = msg.source["content"].get("url")
        if url is not None:
            return [
                LegacyAttachment(
                    url=await self.session.matrix.mxc_to_http(url),
                    legacy_file_id=quote(url),
                    name=msg.source["content"].get("filename") or get_body(msg) or None,
                )
            ]

        return []

    async def __get_encrypted_attachments(self, msg: nio.RoomEncryptedMedia):
        resp = await self.session.matrix.download(mxc=msg.url, filename=None)
        if isinstance(resp, nio.DownloadError):
            self.log.warning(
                "Failed to download an encrypted media in %s: %s", msg, resp
            )
            self.send_text(
                f"/me sent an encrypted media but matridge failed to download it: {resp.message}"
            )
            return []
        else:
            media_data = resp.body
            decrypted_data = nio.crypto.attachments.decrypt_attachment(
                media_data,
                msg.source["content"]["file"]["key"]["k"],
                msg.source["content"]["file"]["hashes"]["sha256"],
                msg.source["content"]["file"]["iv"],
            )
            return [
                LegacyAttachment(
                    data=bytes(decrypted_data),
                    legacy_file_id=quote(msg.url),
                    name=get_body(msg) or None,
                )
            ]

    async def send_matrix_message(
        self,
        msg: nio.RoomMessage,
        replace=None,
        replacement_event_id=None,
        archive_only=False,
    ):
        self.log.debug("Message: %s", msg.source)

        if id_and_new := get_new_message(msg):
            replace, new = id_and_new
            return await self.send_matrix_message(
                new, replace, msg.event_id, archive_only
            )

        kwargs = dict(
            archive_only=archive_only,
            when=server_timestamp_to_datetime(msg),
            correction_event_id=replacement_event_id,
        )
        await self.__add_reply_to(msg, replace, kwargs)
        attachments = await self.__get_attachments(msg, **kwargs)

        await self.send_files(
            attachments,
            msg.event_id,
            body=None if attachments else get_body(msg),
            thread=get_rel(msg.source, "m.thread") or msg.event_id,
            **kwargs,
        )


def strip_reply_fallback(formatted_body: str) -> str:
    obj = bs4.BeautifulSoup(formatted_body, "html.parser")
    if mx_reply := obj.find("mx-reply"):
        if isinstance(mx_reply, bs4.Tag):
            mx_reply.decompose()
    return str(obj.text)


def get_reply_to(source: dict) -> Optional[str]:
    return (
        source.get("content", {})
        .get("m.relates_to", {})
        .get("m.in_reply_to", {})
        .get("event_id")
    )


def get_replace(source: dict):
    return get_rel(source, "m.replace")


def get_rel(source: dict, rel_type: str) -> Optional[str]:
    content = source.get("content")
    if not content:
        return None
    relates_to = content.get("m.relates_to")
    if not relates_to:
        return None
    if relates_to.get("rel_type") != rel_type:
        return None
    return relates_to.get("event_id")


def get_new_content(source: dict) -> Optional[nio.RoomMessage]:
    content = source.get("content")
    if not content:
        return None
    new_content = content.get("m.new_content")
    return new_content


def get_new_message(msg: nio.RoomMessage):
    replace = get_rel(msg.source, "m.replace")
    if not replace:
        return
    return replace, nio.RoomMessage.parse_event(
        {
            "content": get_new_content(msg.source),
            "origin_server_ts": msg.server_timestamp,
            "sender": msg.sender,
            "event_id": replace,
        }
    )


def get_body(msg: nio.RoomMessage):
    if (
        isinstance(msg, nio.RoomMessageFormatted)
        and msg.format == "org.matrix.custom.html"
    ):
        relates_to = msg.source.get("content", {}).get("m.relates_to", {})
        if relates_to.get("rel_type") == "m.replace" or relates_to.get("m.in_reply_to"):
            body = strip_reply_fallback(msg.formatted_body)
        else:
            body = msg.body
    else:
        body = getattr(msg, "body", "")

    if isinstance(msg, nio.RoomMessageEmote):
        body = "/me " + body

    return body


def server_timestamp_to_datetime(event: nio.Event):
    return datetime.fromtimestamp(event.server_timestamp / 1000, tz=timezone.utc)


def get_content(text: str, mentions: Optional[list[Mention]] = None):
    if text.startswith("/me "):
        text = text.removeprefix("/me ").lstrip()
        content = {"msgtype": "m.emote", "body": text}
    else:
        content = {"msgtype": "m.text", "body": text}
    if MESSAGE_STYLING:
        formatted_body = format_for_matrix(
            text, [(m.contact.legacy_id, m.start, m.end) for m in mentions or []]
        )
        content["formatted_body"] = formatted_body
        content["format"] = "org.matrix.custom.html"
    if mentions:
        content["m.mentions"] = {
            "user_ids": [m.contact.legacy_id for m in mentions]  # type:ignore
        }
    return content


log = logging.getLogger()


class Credentials(TypedDict):
    homeserver: str
    user_id: str
    device_id: str
    access_token: str



from collections import defaultdict
import logging

import nio
import pytest

from matridge.reactions import ReactionCache, purge_old_messages


class MockMatrix:
    class session:
        log = logging.getLogger()

    @staticmethod
    async def room_get_event_relations(room, event_id, rel_type):
        if room == "bad":
            return

        yield nio.ReactionEvent(
            reacts_to="msg_id",
            key="<3",
            source=mock_source(
                {
                    "event_id": "1",
                    "sender": "someone",
                    "content": {"m.relates_to": {"event_id": "msg_id", "key": "<3"}},
                }
            ),
        )

        yield nio.ReactionEvent(
            reacts_to="msg_id",
            key="+1",
            source=mock_source(
                {
                    "event_id": "2",
                    "sender": "someone",
                    "content": {"m.relates_to": {"event_id": "msg_id", "key": "+1"}},
                }
            ),
        )
        yield nio.ReactionEvent(
            reacts_to="other_msg_id",
            key="-1",
            source=mock_source(
                {
                    "event_id": "3",
                    "sender": "someone",
                    "content": {
                        "m.relates_to": {
                            "event_id": "other_msg_id",
                            "key": "-1",
                        }
                    },
                }
            ),
        )

    @staticmethod
    async def sync(sync_filter):
        return nio.SyncResponse(None, None, None, None, None, None)

    @staticmethod
    async def get_original_id(room, event_id):
        return event_id


def mock_source(data):
    x = defaultdict(str)
    x.update(data)
    return x


@pytest.mark.asyncio
async def test_no_response():
    cache = ReactionCache(MockMatrix)
    with cache.session() as s:
        assert await cache.get(s, "bad", "msg_id", "sender") == set()


@pytest.mark.asyncio
async def test_remove_unknown_event():
    cache = ReactionCache(MockMatrix)
    with cache.session() as s:
        cache.remove(s, "bad", "unknown")


@pytest.mark.asyncio
async def test_fetch():
    cache = ReactionCache(MockMatrix)
    with cache.session() as s:
        assert await cache.get(s, "good", "msg_id", "someone") == {"<3", "+1"}


@pytest.mark.asyncio
async def test_fetch_add_remove():
    cache = ReactionCache(MockMatrix)
    target = "good", "msg_id", "someone"
    with cache.session() as s:
        await cache.add(s, *target, "prout", "4")
        assert await cache.get(s, *target) == {"<3", "+1", "prout"}
        cache.remove(s, "good", "2")
        assert await cache.get(s, *target) == {"<3", "prout"}


@pytest.mark.asyncio
async def test_fetch_add_remove():
    cache = ReactionCache(MockMatrix)
    with cache.session() as s:
        start = await cache.get(s, "good", "msg_id", "someone")
        for i in range(10):
            await cache.add(s, "good", "msg_id", "someone", str(i), f"whatevs{i}")
        s.commit()

    purge_old_messages(10)
    with cache.session() as s:
        for i in range(10):
            await cache.add(s, str(i), str(i), "someone", "emoji", f"whatevs2-{i}")

        expected = set(str(x) for x in range(10)) | start

        assert await cache.get(s, "good", "msg_id", "someone") == expected
        s.commit()

    purge_old_messages(10)
    with cache.session() as s:
        # triggers a refetch which only returns the
        assert await cache.get(s, "good", "msg_id", "someone") == start
